#include "SSRC_Reader.h"
#include "Writer.h"
#include "Destination.h"

#include "RTP_Packet.h"
#include "RTCP_Packet.h"

#include "ace/Reactor.h"
#include "ace/Timer_Queue.h"

#include <sys/stat.h>
#include <ctype.h>
#include <string.h>

#define USE_STDIO_BUFFERING
/*#define NO_MMFS*/

#ifndef _AIX
#define NO_MMFS
#endif

#ifndef NO_MMFS
#include "TigerShark.h"
#endif

extern double gettimeofday();

SSRC_Reader::SSRC_Reader(char *path, Writer *writer, ACE_Reactor *reactor,
			 int n_packet_types, int *packet_types) :
			 n_packet_types_(n_packet_types),
			 packet_types_(packet_types),
			 writer_(writer), reactor_(reactor),
			 reader_done_(0), ptype_(-1),
			 ssrc_(0),
			 found_rtp_packet_(0),
			 absolute_starting_timestamp_(0),
			 rtcp_interval_(10, 0)
{
    int i;
    printf("%x: create reader for %s\n", this, path);

    strcpy(path_, path);
    
    for (i = 0; i < MaxTimers; i++)
    {
	timer_ids_[i] = 0;
    }

    open_file();
    if (fp_ == NULL)
    {
	perror("SSRC_Reader(): Cannot open file");
	exit(1);
    }

    read_header();
    read_info_file();

    /*
     * Compute some of the synchronization constants.
     */

    Nstart_ = Nsynch_ - (Rsynch_ - Rstart_) / F_;
    printf("Nstart=%f Nsynch=%f Rstart=%f Rsynch=%f F=%f\n",
	   Nstart_, Nsynch_, Rstart_, Rsynch_, F_);

}

/*
 * Transform an RTP timestamp to the NTP timestamp.
 * The returned timestamp is the absolute time corresponding to
 * this timestamp.
 */
double SSRC_Reader::rtp_to_ntp(unsigned long rtp_ts)
{
    double Ri, Ni;

    Ri = (double) rtp_ts;

    Ni = (Ri - Rsynch_) / F_ + Nsynch_;

    return Ni;
}

/*
 * Transform an RTP timestamp to the NTP timestamp.
 * Timestamp returned is relative to the start of the recording.
 */
double SSRC_Reader::rtp_to_rel_ntp(unsigned long rtp_ts)
{
    double Ni;

    Ni = rtp_to_ntp(rtp_ts);
    Ni = Ni - Nstart_;

    return Ni;
}

void SSRC_Reader::open_file()
{
    struct stat s;
    
    if (strncmp(path_, "/Net/vs01", 9) == 0)
    {
	sprintf(path_, "/mmfs%s", path_ + 9);
    }

    fp_ = fopen(path_, "r");

    if (fp_ == (FILE *) NULL)
    {
	perror("fopen");
	exit(1);
    }

    if (fstat(fileno(fp_), &s) < 0)
    {
	perror("fstat");
	return;
    }

#ifdef USE_STDIO_BUFFERING
    printf("Setting stdio blocksize to %d\n", s.st_blksize);
    fp_buffer_ = new char[s.st_blksize];

    if (setvbuf(fp_, fp_buffer_, _IOFBF, s.st_blksize) != 0)
    {
	perror("setvbuf");
    }
#else
    fp_buffer_ = NULL;
    printf("Disabling stdio buffering\n");
    if (setvbuf(fp_, NULL, _IONBF, 0) != 0)
    {
	perror("setvbuf");
    }
#endif
    
#ifndef NO_MMFS
    if (s.st_vfstype == VFSTYPE_MMFS)
    {
	int ival;

	ival = 5000000;

#ifdef OLD_MMFS
	
	if (mmfs_qioctl(fileno(fp_), SET_RATE, &ival) < 0)
	    perror("SET_RATE");

	if (mmfs_qioctl(fileno(fp_), GET_RATE, &ival) < 0)
	    perror("GET_RATE");
	else
	    printf("rate=%d\n", ival);

	ival = TS_BUFFERS;
	if (mmfs_qioctl(fileno(fp_), SET_BUFFERING, &ival) < 0)
	    perror("SET_BUFFERING");
#else
	if (tsfattr(fileno(fp_), SET_RATE, &ival, NULL) < 0)
	    perror("SET_RATE");

	if (tsfattr(fileno(fp_), GET_RATE, &ival, NULL) < 0)
	    perror("GET_RATE");
	else
	    printf("rate=%d\n", ival);

	ival = TS_BUFFERS;
	if (tsfattr(fileno(fp_), SET_BUFFERING, &ival, NULL) < 0)
	    perror("SET_BUFFERING");
#endif
	
    }
    else
    {
	printf("Warning: %s not in MMFS\n", path_);
	return;
    }
#endif
}

void SSRC_Reader::read_info_file()
{
    FILE *info_fp;
    char ipath[1024];
    char line[1024], *s, *t;
    
    sprintf(ipath, "%s_info", path_);
    info_fp = fopen(ipath, "r");

    if (!info_fp)
    {
	printf("Couldn't open info file %s\n", ipath);
	exit(1);
    }

    while (fgets(line, sizeof(line), info_fp) != NULL)
    {
	s = strchr(line, ':');
	if (s)
	{
	    *s = '\0';
	    s++;
	    while (isspace(*s))
		s++;

	    t = s + 1;
	    while (!isspace(*t))
		t++;
	    *t = '\0';

	    handle_info_item(line, s);
	}
    }
    fclose(info_fp);
}

/*
 * Handle the metadata that we're interested in.
 */
void SSRC_Reader::handle_info_item(char *tag, char *value)
{
    printf("GOT info <%s> <%s>\n", tag, value);

#define FMATCH(s) if (strcmp(tag, #s) == 0)
#define MATCH(s) else FMATCH(s)


    FMATCH(starting_timestamp)
    {
	starting_timestamp_ = atoi(value);
	Rstart_ = (double) starting_timestamp_;
    }
    MATCH(packet_types)
    {
	if (strchr(value, ' '))
	{
	    printf("Cannot handle streams with multiple packet types\n");
	    exit(1);
	}
	ptype_ = atoi(value);
	
	switch (ptype_)
	{
	case 0:		/* PCMU */
	case 1:		/* 1016 */
	case 2:		/* G721 */
	case 3:		/* GSM */
	case 4:		/* unassigned */
	case 5:		/* DVI4 */
	case 7:		/* LPC */
	case 8:		/* PCMA */
	case 9:		/* G722 */
	case 15:		/* G728 */
		
	    timestamp_frequency_ = 8000;
	    break;

	case 6:		/* DVI4 */
	    timestamp_frequency_ = 16000;
	    break;

	case 10:		/* L16 stereo */
	case 11:		/* L16 mono  */
	    timestamp_frequency_ = 44100;
	    break;

	case 14:		/* MPA */
	case 25:		/* CelB */
	case 26:		/* JPEG */
	case 28:		/* nv */
	case 31:		/* H261 */
	case 32:		/* MPV */
	case 33:		/* MP2T */
	    timestamp_frequency_ = 90000;
	    break;

	default:
	  printf("dunno about type %d yet\n", ptype_);
	  exit(1);
	}
    }
    MATCH(synch_ntp_timestamp)
    {
	Nsynch_ = synch_ntp_timestamp_ = atof(value);
    }
    MATCH(synch_rtp_timestamp)
    {
	synch_rtp_timestamp_ = atoi(value);
	Rsynch_ = (double) synch_rtp_timestamp_;
    }

    F_ = (double) timestamp_frequency_;
}

int SSRC_Reader::read_header()
{
    if (fread((void *) &header_, sizeof(header_), 1, fp_) != 1)
    {
	perror("SSRC_Reader::read_header(): fread failed");
	return 0;
    }

    /*
     * We will compute a better value for this later, from
     * the rtp timestamp from the first data packet and the
     * ntp/rtp synch information.
     */

    start_time_.sec(ntohl(header_.start_s));
    start_time_.usec(ntohl(header_.start_us));
		    
    original_source_.set(ntohs(header_.port), ntohl(header_.source));

    time_t thetime = (time_t) start_time_.sec();
    printf("Session was started by %s/%d at %d.%06d %s",
	   original_source_.get_host_name(),
	   original_source_.get_port_number(),
	   start_time_.sec(), start_time_.usec(),
	   ctime(&thetime));

    return 1;
}

/*
 * Compute the playout delay for the packet currently buffered
 * in packet_/packet_buffer_.
 */
double SSRC_Reader::compute_playout_delay()
{
    ACE_Time_Value delay;
//    timeval cur_time;

//    ::gettimeofday(&cur_time);
//    ACE_Time_Value cur = ACE_Timer_Queue::current_time();

    double cur_d = gettimeofday();

/*
    printf("compute_delay: offset=%d.%06d ms=%d.%06d cur=%d.%06d\n",
	   packet_offset_.sec(), packet_offset_.usec(),
	   mapped_start_.sec(), mapped_start_.usec(),
	   cur.sec(), cur.usec());
*/

    /*
    delay = packet_offset_ + mapped_start_ - cur;
    */


//    delay = ACE_Time_Value(0, C1_ * packet_timestamp_) + C3_ - cur;

/*
 * THIS IS THE OLD "WORKING" PLAYOUT COMPUTATION

    double lt = ((double) (packet_timestamp_ - starting_timestamp_)) / (double) timestamp_frequency_;


    double delay_d = playback_start_d_ - cur_d + lt;
*/
    
//    delay = writer_->playback_start() - cur + ACE_Time_Value((packet_timestamp_ - starting_timestamp_) / timestamp_frequency_);


/*
   printf("compute_delay: ts=%d cur=%f delay=%f %d %d\n",
	   packet_timestamp_, cur_d, delay_d, lts, ltus);
*/

    /*
     * Playout time for this packet is the relative ntp timestamp
     * of the RTP timestamp, plus the playback start time. Playback
     * delay is that time minus the current time.
     */

    double rel = rtp_to_rel_ntp(packet_timestamp_);
    double delay_d = rel + playback_start_d_ - cur_d;
#if 0
    printf("Compute delay: ts=%u rel=%f start=%f cur=%f delay=%f\n",
	   packet_timestamp_, rel, playback_start_d_, cur_d, delay_d);
#endif

    return delay_d;
}

/*
 * Select the destinations that we will send to by matching the packet
 * types.
 */
void SSRC_Reader::select_destinations(int n_dests, Destination **dests)
{
    int i, j;

    n_destinations_ = 0;
    destinations_ = new Destination*[n_dests];

    for (i = 0; i < n_dests; i++)
    {
	for (j = 0; j < n_packet_types_; j++)
	{
	    if (dests[i]->wants_type(packet_types_[j]))
	    {
		printf("Reader %x selects destination %x for type %d\n",
		       this, dests[i], packet_types_[j]);
		destinations_[n_destinations_++] = dests[i];
	    }
	}
    }
}
 

/*
 * Start this whole production by reading the next packet, calculating the
 * delay until it needs to be sent, and scheduling a timeout.
 *
 * If the time is too short, go ahead and send the backet now.
 */

void SSRC_Reader::start(ACE_Time_Value &playback_start,
			ACE_Time_Value &base_offset)
{
/*
    if (am_leader())
    {
	stream_started(this);
	timer_ids_[Counter] = reactor_->schedule_timer(this, 0, 1, 1);
    }
    else
    {
	leader_->stream_started(this);
    }
*/
    
    mapped_start_ = start_time_ + base_offset;

    playback_start_d_ = (double) playback_start.sec() +
      (double) playback_start.usec() / 1000000.0;

#if 0
    /*
     * Jettison this, now that we read the _info file.
     */
    
    /*
     * Read packets until we get the first RTP packet; then compute
     * the offset in terms of rtp timestamp units.
     */

    found_rtp_packet_ = 0;
    while (!found_rtp_packet_)
    {
	if (!read_packet())
	{
	    done_sending();
	    return;
	}
    }
#endif
    
    printf("Started up, initial timestamp is %u freq is %u\n", starting_timestamp_,
	   timestamp_frequency_);

    /*
     * Compute our useful constants
     */

    C1_ = 1e6 / timestamp_frequency_;
    C2_ = 1e6 * starting_timestamp_ / timestamp_frequency_;
    C3_ = playback_start - ACE_Time_Value(0, C2_);

    printf("Got C1=%u C2=%u C3=%d.%06d\n", C1_, C2_,
	   C3_.sec(), C3_.usec());

    /*
     * Start up the timer for the RTCP reports
     */
    timer_ids_[RTCP_Report] = reactor_->schedule_timer(this, 0, 0,
						       rtcp_interval_);

    read_and_send();
}

void SSRC_Reader::read_and_send()
{
    int first = 0;

    send_packet();
    while (1)
    {
	if (!read_packet())
	{
	    done_sending();
	    return;
	}

	/*
	 * We just spin sending out RTCP packets for now
	 */
	if (packet_.type == 0)	// RTP packet
	{
	    if (packet_timestamp_ == last_packet_timestamp_)
	    {
//		printf("dup! %u %u\n", packet_timestamp_, last_packet_timestamp_);
		send_packet();
		continue;
	    }
	    else
	    {
//		printf("nodup! %u %u\n", packet_timestamp_, last_packet_timestamp_);
	    }

	    last_packet_timestamp_ = packet_timestamp_;

	    double playout_delay = compute_playout_delay();
	    
	    if (playout_delay > 0.0)
	    {
		/*
		 * Only schedule a timeout if the delay is greater than 10ms
		 */
		if (playout_delay > 0.010)
		{
		    long s, us;
		    s = (long) playout_delay;
		    us = (long) ((playout_delay - (double) s) * 1000000.0);
		    timer_ids_[Update] =
			reactor_->schedule_timer(this, (void *) Update,
						 ACE_Time_Value(s, us), 0);
		    break;
		}
		else		// Just send it out now
		{
		    send_packet();
		}
	    }
	    else
	    {
		printf("OUCH %x delay=%f offset=%d\n",
		       this, playout_delay, packet_.offset);
	    }
	}
	else			// RTCP packet
	{
	    send_packet();
	}
    }
}

void SSRC_Reader::done_sending()
{
    int i;

    reader_done_ = 1;

    printf("reader %x done\n", this);

    writer_->reader_finished(this);
/*
    if (am_leader())
    {
	stream_finished(this);
    }
    else
    {
	leader_->stream_finished(this);
    }
*/
}

int SSRC_Reader::read_packet()
{
    RTP_Packet::RTP_Header *rp;
    RTCP_Packet::RTCP_Header *rtcp;
    
    if (fread((void *) &packet_, sizeof(packet_), 1, fp_) != 1)
    {
	perror("SSRC_Reader::read_packet(): fread failed");
	return 0;
    }

    packet_.length = ntohs(packet_.length);
    packet_.type = ntohs(packet_.type);
    packet_.offset = ntohl(packet_.offset);

/*
    printf("pkt: l=%d t=%d o=%d\n", packet_.length, packet_.type,
	   packet_.offset);
*/

    packet_offset_.sec(packet_.offset / 1000);
    packet_offset_.usec((packet_.offset % 1000) * 1000);

    packet_buffer_length_ = packet_.length - sizeof(packet_);
    if (fread(packet_buffer_, packet_buffer_length_, 1, fp_) != 1)
    {
	perror("SSRC_Reader::read_packet(): fread failed");
	return 0;
    }

    if (packet_.type == 0)
    {
	rp = (RTP_Packet::RTP_Header *) packet_buffer_;
	packet_timestamp_ = rp->get_timestamp();

/*
	printf("rp v=%d type=%d ts=%d ssrc=%x\n", rp->get_version(),
	       rp->get_packet_type(), packet_timestamp_, rp->get_ssrc());
*/

	if (ssrc_ == 0)
	{
	    ssrc_ = rp->get_ssrc();
	}

	
#if 0
	/*
	 * Don't worry about this; we get this info from the
	 * _info file now.
	 */
	if (!found_rtp_packet_)
	{
	    found_rtp_packet_ = 1;
	    ptype_ = rp->get_packet_type();

	    last_packet_timestamp_ = starting_timestamp_ = packet_timestamp_;
	    if (absolute_starting_timestamp_ == 0)
		absolute_starting_timestamp_ = packet_timestamp_;

	    /*
	     * Sampling frequency assignments from RFC 1890
	     * (http://www.internic.net/rfc/rfc1890.txt)
	     */

	    switch (rp->get_packet_type())
	    {
	    case 0:		/* PCMU */
	    case 1:		/* 1016 */
	    case 2:		/* G721 */
	    case 3:		/* GSM */
	    case 4:		/* unassigned */
	    case 5:		/* DVI4 */
	    case 7:		/* LPC */
	    case 8:		/* PCMA */
	    case 9:		/* G722 */
	    case 15:		/* G728 */
		
		timestamp_frequency_ = 8000;
		break;

	    case 6:		/* DVI4 */
		timestamp_frequency_ = 16000;
		break;

	    case 10:		/* L16 stereo */
	    case 11:		/* L16 mono  */
		timestamp_frequency_ = 44100;
		break;

//	    case 26:
//
//		Where'd this come from?? From RFC1890 it's 90khz
//		timestamp_frequency_ = 65536;
//		break;

	    case 14:		/* MPA */
	    case 25:		/* CelB */
	    case 26:		/* JPEG */
	    case 28:		/* nv */
	    case 31:		/* H261 */
	    case 32:		/* MPV */
	    case 33:		/* MP2T */
		timestamp_frequency_ = 90000;
		break;

	    default:
		printf("dunno about type %d yet\n", rp->get_packet_type());
		exit(1);
	    }
	}
#endif
    }
    else
    {
	rtcp = (RTCP_Packet::RTCP_Header *) packet_buffer_;
	if (ssrc_ == 0)
	{
	    ssrc_ = rtcp->get_ssrc();
	}
//	packet_timestamp_ = rtcp->get_timestamp();
    }

    return 1;
}

void SSRC_Reader::send_packet()
{
    int i;

    if (packet_.type == 0)
    {
//      printf("%x send packet\n", ssrc_);
	RTP_Packet::RTP_Header *rp = (RTP_Packet::RTP_Header *) packet_buffer_;

	for (i = 0; i < n_destinations_; i++)
	{
	    destinations_[i]->send_rtp_packet(rp->get_packet_type(),
				     packet_buffer_,
				     packet_buffer_length_);
	}
    }
    else
    {
	for (i = 0; i < n_destinations_; i++)
	{
	    destinations_[i]->send_rtcp_packet(packet_buffer_,
					       packet_buffer_length_);
	}
    }
}

void SSRC_Reader::stop()
{
    reactor_->cancel_timer(timer_ids_[Update]);
    timer_ids_[Update] = 0;

/*
    if (am_leader())
    {
	reactor_->cancel_timer(timer_ids_[Counter]);
	timer_ids_[Counter] = 0;
    }
*/
}

void SSRC_Reader::reset()
{
}


int SSRC_Reader::handle_timeout(const ACE_Time_Value &tv, const void *arg)
{
    switch ((int) arg)
    {
    case Update:
    /*
     * O-tay, we're ready to send the packet we've got saved in our
     * hot little hands (er, object, anyway).
     *
     * Send it out, and recompute the delay until the next playout.
     */

	/*
	 * Don't call send_packet here, since the first
	 * thing that read_and_send does is a send_packet().
	 * Gee, think this is the source of the dups?
	send_packet();
	 */
	read_and_send();
	break;

    case Counter:
	send_counter_update();
	break;

    case RTCP_Report:
	printf("RTCP Report time\n");
	break;
    }
    return 0;
}

void SSRC_Reader::send_counter_update()
{
    int i;
    double clock;
    char buf[100];

    clock = ((double) packet_timestamp_ - starting_timestamp_) /
	(double) timestamp_frequency_;
    sprintf(buf, "%x %f", ssrc_, clock);

    for (i = 0; i < n_destinations_; i++)
    {
	destinations_[i]->send_callback("counter", buf);
    }
}

void SSRC_Reader::set_leader(void)
{
    leader_ = (SSRC_Reader *) NULL;
    n_streams_started_ = 0;
}

void SSRC_Reader::set_leader(SSRC_Reader *l)
{
    leader_ = l;
    n_streams_started_ = -1;
}

void SSRC_Reader::stream_started(SSRC_Reader *s)
{
    if (!am_leader())
    {
	printf("Ack! non-leader got stream_started()\n");
	exit(1);
    }

    n_streams_started_++;

    printf("leader got stream started, nstarted=%d\n", n_streams_started_);
}

void SSRC_Reader::stream_finished(SSRC_Reader *s)
{
    if (!am_leader())
    {
	printf("Ack! non-leader got stream_started()\n");
	exit(1);
    }

    n_streams_started_--;

    printf("leader got stream finished, nstarted=%d\n", n_streams_started_);

    if (n_streams_started_ == 0)
    {
	int i;
	
	reactor_->cancel_timer(timer_ids_[Counter]);
	timer_ids_[Counter] = 0;

	for (i = 0; i < n_destinations_; i++)
	{
	    destinations_[i]->send_callback("stream_done", "all");
	}
	
	printf("All streams done\n");
    }
}

