#include "SSRC_Writer.h"
#include "ntp.h"

#include "ace/Time_Value.h"
#include "ace/Timer_Queue.h"
#include "ace/INET_Addr.h"

#include "Disk_Writer.h"
#include <stdio.h>
#include <sys/stat.h>

#define USE_STDIO_BUFFERING
/*#define NO_MMFS*/

#ifndef _AIX
#define NO_MMFS
#endif

#ifndef NO_MMFS
#include "TigerShark.h"
#endif

#ifdef USE_IBM_CL
long const & key(SSRC_Writer * const s)
{
    return s->ssrc();
}
#endif

SSRC_Writer::SSRC_Writer(Disk_Writer *DW, long ssrc) :
			 disk_writer_(DW),
			 ssrc_(ssrc),
			 last_rtp_seq(0),
			 bytes_(0),
			 fp_(NULL),
			 info_fp_(NULL),
			 writing_output_(0),
			 have_saved_rtcp_packet_(0),
			 missed_packets_(0),
			 packets_seen_(0),
			 have_synch_info_(0)
{
    printf("New ssrc %x\n", ssrc_);

    if (disk_writer_->directory())
    {
	sprintf(filename_, "%s/%x", disk_writer_->directory(), ssrc_);

	open_file();
    }
    bzero(types_seen_, sizeof(types_seen_));
}

SSRC_Writer::~SSRC_Writer()
{
    terminate();
    printf("SW(%x) Destruct\n", ssrc_);
}

void SSRC_Writer::terminate()
{
    printf("SW(%x) got terminate\n", ssrc_);
    fclose(fp_);
    fp_ = NULL;

    write_final_info();
}

/*
 * Spit out info we know about the start of the stream.
 *
 * SSRC
 * starting time
 */
void SSRC_Writer::write_initial_info()
{
    ACE_Time_Value duration;

    if (!disk_writer_->directory())
	return;

    sprintf(info_filename_, "%s/%x_info", disk_writer_->directory(), ssrc_);
    info_fp_ = fopen(info_filename_, "w");
    if (info_fp_ == (FILE *) NULL)
    {
	perror("info filename open");
	exit(1);
    }
    fprintf(info_fp_, "ssrc: %x\n", ssrc_);

    fprintf(info_fp_, "start_time: %d.%06d\n",
	    start_time_.sec(), start_time_.usec());

    fflush(info_fp_);
}

/*
 * Spit out the end-of-stream summary information:
 *
 * Duration
 */
void SSRC_Writer::write_final_info()
{
    ACE_Time_Value duration;

    if (info_fp_ == (FILE *) NULL)
	return;
    
    duration = ACE_OS::gettimeofday() - start_time_;
    
    fprintf(info_fp_, "packet_types:");
    for (int i = 0; i < 128; i++)
    {
	if (types_seen_[i])
	    fprintf(info_fp_, " %d", i);
    }
    fprintf(info_fp_, "\n");

    fprintf(info_fp_, "duration: %d.%06d\n", duration.sec(), duration.usec());

    fclose(info_fp_);
}


void SSRC_Writer::open_file()
{
    struct stat s;

    fp_ = fopen(filename_, "w");

    if (fp_ == (FILE *) NULL)
    {
	perror("SSRC_Writer::open_file open:");
	exit(1);
    }

    if (fstat(fileno(fp_), &s) < 0)
    {
	perror("fstat");
	return;
    }

#ifdef USE_STDIO_BUFFERING
    printf("Setting stdio blocksize to %d\n", s.st_blksize);
    fp_buffer_ = new char[s.st_blksize];

    if (setvbuf(fp_, fp_buffer_, _IOFBF, s.st_blksize) != 0)
    {
	perror("setvbuf");
    }
#else
    fp_buffer_ = NULL;
    printf("Disabling stdio buffering\n");
    if (setvbuf(fp_, NULL, _IONBF, 0) != 0)
    {
	perror("setvbuf");
    }
#endif
    
#ifndef NO_MMFS
    if (s.st_vfstype == VFSTYPE_MMFS)
    {
	int ival;

	ival = 5000000;

#ifdef OLD_MMFS
	
	if (mmfs_qioctl(fileno(fp_), SET_RATE, &ival) < 0)
	    perror("SET_RATE");

	if (mmfs_qioctl(fileno(fp_), GET_RATE, &ival) < 0)
	    perror("GET_RATE");
	else
	    printf("rate=%d\n", ival);

	ival = TS_BUFFERS;
	if (mmfs_qioctl(fileno(fp_), SET_BUFFERING, &ival) < 0)
	    perror("SET_BUFFERING");
#else
	if (tsfattr(fileno(fp_), SET_RATE, &ival, NULL) < 0)
	    perror("SET_RATE");

	if (tsfattr(fileno(fp_), GET_RATE, &ival, NULL) < 0)
	    perror("GET_RATE");
	else
	    printf("rate=%d\n", ival);

	ival = TS_BUFFERS;
	if (tsfattr(fileno(fp_), SET_BUFFERING, &ival, NULL) < 0)
	    perror("SET_BUFFERING");
#endif
	
    }
    else
    {
	printf("Warning: %s not in MMFS\n", filename_);
	return;
    }
#endif
}

void SSRC_Writer::handle_packet(RTP_Packet::RTP_Header *packet_header,
				char *packet, int packet_size,
				ACE_INET_Addr &packet_source,
				int initial_packet)
{
    SSRC_Writer *swriter;

    /*
     * If we aren't writing output yet, we need to start.
     * Open the file, write the header, and write the saved
     * RTCP packet if there is one.
     */
	 
    if (!writing_output_)
    {
	write_file_header(packet_source);
	write_initial_info();
	if (have_saved_rtcp_packet_)
	{
	    int len;
	    char *buf;
	    
	    buf = saved_rtcp_packet_.get_buffer(&len);
	    write_packet(buf, saved_rtcp_packet_size_, 1);
	}

	/*
	 * Write out the timestamp for this first RTP packet in the
	 * file.
	 */
	fprintf(info_fp_, "starting_timestamp: %u\n",
		packet_header->get_timestamp());
	fflush(info_fp_);
	writing_output_ = 1;
    }

    bytes_ += packet_size;

    packets_seen_++;
    last_rtp_seq++;
    if (packet_header->get_sequence() != last_rtp_seq &&
	last_rtp_seq != 65536)
    {
#ifdef VERBOSE
	printf("SW(%x): missed packet %d\n", ssrc_, last_rtp_seq);
#endif
	missed_packets_++;
	last_rtp_seq = packet_header->get_sequence();
    }

    write_packet(packet, packet_size, 0);

    int packet_type = packet_header->get_packet_type();
    if (!types_seen_[packet_type])
    {
	types_seen_[packet_type] = 1;
	if (info_fp_)
	{
	    fprintf(info_fp_, "packet_types_seen+ %d\n", packet_type);
	    fflush(info_fp_);
	}
    }

    /* 
    printf("SW(%x): got rtp packet of length %d\n", ssrc_, packet_size);

    printf("flags=%x v=%d pad=%d ext=%d cc=%d marker=%d pt=%d seq=%d ssrc=%x ts=%u\n",
	   
	   packet_header->get_flags(),
	   packet_header->get_version(),
	   packet_header->get_padding(),
	   packet_header->get_extension(),
	   packet_header->get_csrc_count(),
	   packet_header->get_marker(),
	   packet_header->get_packet_type(),
	   packet_header->get_sequence(),
	   packet_header->get_ssrc(),
	   packet_header->get_timestamp());
       */
}
    
void SSRC_Writer::handle_packet(RTCP_Packet::RTCP_Header *packet_header,
				char *packet, int packet_size,
				ACE_INET_Addr &packet_source,
				int initial_packet)
{
    if (!writing_output_)
    {
	char *buf;
	int len;
	buf = saved_rtcp_packet_.get_buffer(&len);
	bcopy(packet, buf, packet_size);
	saved_rtcp_packet_size_ = packet_size;
	have_saved_rtcp_packet_ = 1;
    }
    else
    {
	write_packet(packet, packet_size, 1);
    }

}

void SSRC_Writer::handle_subpacket(RTCP_Packet::RTCP_Header *packet_header,
				char *packet, int packet_size,
				ACE_INET_Addr &packet_source)
{
    switch (packet_header->get_packet_type())
    {
    case RTCP_Packet::PT_BYE:
    case RTCP_Packet::PT_BYE_OLD:	
	disk_writer_->ssrc_finished(this);
	break;
	
    case RTCP_Packet::PT_SR:
    case RTCP_Packet::PT_SR_OLD:

	/*
	 * If we aren't writing output yet, remember the ssrc &
	 * timestamp for this SR. We will use this on playback
	 * to precisely recreate the timing of the original stream
	 * (by remembering the mapping between wallclock time and
	 * rtp timestamp).
	 */
	if (!have_synch_info_)
	{
	    RTCP_Packet::Sender_Info *si;

	    si = (RTCP_Packet::Sender_Info *) (packet + 8);

	    double d = NTP_TO_GETTIMEOFDAY(ntohl(si->ntp_timestamp_msw_),
					   ntohl(si->ntp_timestamp_lsw_));
	    
	    printf("SI: ssrc=%x ntp %u %u d=%f rtp=%u pack=%d oct=%d\n",
		   packet_header->get_ssrc(),
		   ntohl(si->ntp_timestamp_msw_),
		   ntohl(si->ntp_timestamp_lsw_),
		   d,
		   ntohl(si->rtp_timestamp_),
		   ntohl(si->packet_count_),
		   ntohl(si->octet_count_));

	    fprintf(info_fp_, "synch_ntp_timestamp: %f\n", d);
	    fprintf(info_fp_, "synch_rtp_timestamp: %u\n",
		    ntohl(si->rtp_timestamp_));
	    
	    have_synch_info_ = 1;

	    break;
	}
    }

    /*
    if (packet_header->get_packet_type() == RTCP_Packet::PT_BYE ||
	packet_header->get_packet_type() == RTCP_Packet::PT_BYE_OLD)
    {
	disk_writer_->ssrc_finished(this);
    }
    */
}

void SSRC_Writer::write_file_header(ACE_INET_Addr &source)
{
    RTP_Dump_Header rd_header_;

    if (!disk_writer_->directory())
	return;

    start_time_ = ACE_OS::gettimeofday();

    printf("SW(%x): write_file_header()\n", ssrc_);

    rd_header_.start_s = htonl(start_time_.sec());
    rd_header_.start_us = htonl(start_time_.usec());
    rd_header_.source = htonl(source.get_ip_address());
    rd_header_.port = htons(source.get_port_number());

    if (fwrite((unsigned char *) &rd_header_, sizeof(rd_header_), 1, fp_) != 1)
    {
	perror("SSRC_Writer::write_file_header: write failed");
	exit(1);
    }

    start_ms_ = start_time_.sec() * 1000 + start_time_.usec() / 1000;
}

void SSRC_Writer::write_packet(char *packet, int length, int type)
{
    ACE_Time_Value now = ACE_OS::gettimeofday();

    if (!disk_writer_->directory())
	return;

    rd_packet_.length = htons(length + sizeof(rd_packet_));
    rd_packet_.type = htons(type);
    rd_packet_.offset = htonl(now.sec() * 1000 + now.usec() / 1000 - start_ms_);
    if (fwrite(&rd_packet_, sizeof(rd_packet_), 1, fp_) != 1)
    {
	perror("SSRC_Writer::write_packet: write header failed");
	exit(1);
    }

    if (fwrite((u_char *) packet, length, 1, fp_) != 1)
    {
	perror("SSRC_Writer::write_packet: write data failed");
	exit(1);
    }
}

void SSRC_Writer::tick(const ACE_Time_Value &this_tick,
		       const ACE_Time_Value &last_tick,
		       const ACE_Time_Value &delta)
{
    double bw;

    bw = 8.0 * (double) bytes_ / ((double) delta.sec() + (double) delta.usec() / 1000000.0);
    printf("%d bytes in %d.%06d sec = %f Mb/s %d/%d missed packets (%5.2f%%)\n",
	   bytes_, delta.sec(), delta.usec(), bw / 1048576.0,
	   missed_packets_, packets_seen_,
	   100 * (double) missed_packets_ / (double) packets_seen_);
    bytes_ = 0;
    missed_packets_ = packets_seen_ = 0;
}
