#include <stdio.h>

#include "Writer.h"
#include "SSRC_Reader.h"
#include "Destination.h"
#include "ace/Timer_Queue.h"
#include "ace/Reactor.h"
#include "RTCP_Packet_Handler.h"

#include "nexus_utils.h"

Writer::Writer(ACE_Reactor *reactor) :
		reactor_(reactor),
		finished_(0),
		running_(0),
		n_listening_ports_(0),
		n_destinations_(0),
		n_destinations_done_(0),
		n_active_clients_(0),
		n_finished_clients_(0)
{
}

int Writer::handle_timeout(const ACE_Time_Value &tv, const void *arg)
{
    send_counter_update();

    return 0;
}

void Writer::send_counter_update(void)
{
    int i;
    double clock;
    char buf[100];

    clock = leader_->absolute_elapsed_time();
    sprintf(buf, "%f", clock);

    for (i = 0; i < n_clients_; i++)
    {
	send_callback(clients_[i], "counter", buf);
    }
}

int Writer::handle_signal (ACE_HANDLE signum HANDLE_SIGNAL_ARGS)
{
    printf("Got signal %d\n", signum);
    finished_ = 1;
    return 0;
}


void Writer::add_client(nexus_global_pointer_t *gp)
{
    clients_[n_clients_++] = gp;

    //
    // XXX
    // Hey! what's a client used for
    //
}

void Writer::add_file(char *path, int n_packet_types, int *packet_types)
{
    printf("Writer::add_file(%s)\n", path);

    SSRC_Reader *reader;

    reader = new SSRC_Reader(path, this, reactor_,
			     n_packet_types, packet_types);

    int was_running = running_;
    if (was_running)
      stop();

    add_reader(reader);

    if (was_running)
      start();
}

void Writer::remove_file(char *path)
{
    printf("Writer::remove_file(%s)\n", path);
}

void Writer::add_destination(char *host, int port,
			     nexus_global_pointer_t *cb_gp,
			     int n_packet_types,
			     int *packet_types)
{
    printf("Writer::add_destination(%s, %d)\n", host, port);

    Destination *dest;

    dest = new Destination(host, port, cb_gp, 
			   n_packet_types, packet_types);

    int was_running = running_;
    if (was_running)
      stop();

    add_destination(dest);
    ensure_listener(port);

    if (was_running)
      start();
}

/*
 * Make sure we have a socket listening on this port.
 */
void Writer::ensure_listener(int port)
{
    int i;

    for (i = 0; i < n_listening_ports_; i++)
    {
	if (listening_ports_[i] == port)
	{
	    return;
	}
    }

    printf("adding listener for %d\n", port);

    ACE_INET_Addr *addr;
    ACE_SOCK_Dgram *sock;
    RTCP_Packet_Handler *handler;

    addr = new ACE_INET_Addr(port + 1);
    sock = new ACE_SOCK_Dgram(*addr);
    handler = new RTCP_Packet_Handler(sock, this);

    printf("created listener on %s %d\n", addr->get_host_addr(), addr->get_port_number());;

    listening_ports_[n_listening_ports_] = port;
    listening_handlers_[n_listening_ports_] = handler;

    n_listening_ports_++;
    printf("registering\n");
    reactor_->register_handler(handler, ACE_Event_Handler::READ_MASK);
    printf("done\n");
}

void Writer::handle_packet(RTCP_Packet::RTCP_Header *packet_header,
			   char *packet, int packet_size,
			   ACE_INET_Addr &packet_source)
{
    int i, found;
    
#if 0
    printf("Wrter: got rtcp packet of length %d\n", packet_size);
    printf("flags=%x version=%d padding=%d rc=%d pt=%d len=%d ssrc=%x\n",
	   packet_header->get_flags(),
	   packet_header->get_version(),
	   packet_header->get_padding(),
	   packet_header->get_reception_count(),
	   packet_header->get_packet_type(),
	   packet_header->get_length(),
	   packet_header->get_ssrc());
#endif

    if (!running_)
    {
	start();
    }

    found = 0;
    for (i = 0; i < n_active_clients_; i++)
    {
	if (active_clients_[i] == packet_header->get_ssrc())
	{
	    found = 1;
	    break;
	}
    }
    if (!found)
    {
	active_clients_[n_active_clients_++] = packet_header->get_ssrc();
	printf("add client ssrc %x\n", packet_header->get_ssrc());
    }
}

void Writer::handle_subpacket(RTCP_Packet::RTCP_Header *packet_header,
				char *packet, int packet_size,
				ACE_INET_Addr &packet_source)
{
/*    printf("got subpacket type %d\n", packet_header->get_packet_type());*/
    
    if (packet_header->get_packet_type() == RTCP_Packet::PT_BYE ||
	packet_header->get_packet_type() == RTCP_Packet::PT_BYE_OLD)
    {

	printf("Got done from %s/%d %x\n", packet_source.get_host_addr(),
	       packet_source.get_port_number(),
	       packet_header->get_ssrc());

	for (int i = 0; i < n_active_clients_; i++)
	{
	    if (active_clients_[i] == packet_header->get_ssrc())
	    {
		printf("done for client %x\n", packet_header->get_ssrc());
		n_finished_clients_++;
	    }
	}

	if (n_finished_clients_ >= n_active_clients_)
	{
	    printf("Last destination left\n");
	    quit();
	}
    }
}

void Writer::remove_destination(char *host, int port)
{
    printf("Writer::remove_destination(%s, %d)\n", host, port);

    Destination *dest;

    dest = find_destination(host, port);

    int was_running = running_;
    if (was_running)
      stop();
    
    remove_destination(dest);

    if (was_running)
      start();
}

void Writer::start()
{
    printf("Writer::start()\n");

    int i;
    ACE_Time_Value min_start;

    if (n_readers_== 0)
    {
	printf("Returning due to n_readers_ = 0\n");
	return;
    }

    running_ = 1;

    for (i = 0; i < n_readers_; i++)
    {
	readers_[i]->select_destinations(n_destinations_, destinations_);
    }

    min_start = readers_[0]->start_time();
    leader_ = readers_[0];
    
    for (i = 1; i < n_readers_; i++)
    {
	if (readers_[i]->start_time() < min_start)
	{
	    min_start = readers_[i]->start_time();
	    leader_ = readers_[i];
	}
    }

    playback_start_ = ACE_OS::gettimeofday();
    base_offset_ = playback_start_ - min_start;

    printf("Min start time is %d.%06d offset=%d.%06d playback_start=%d.%06d\n",
	   min_start.sec(), min_start.usec(),
	   base_offset_.sec(), base_offset_.usec(),
	   playback_start_.sec(), playback_start_.usec());

    counter_id_ = reactor_->schedule_timer(this, 0, 1, 1);
    printf("schedule returns %d\n", counter_id_);

    n_readers_running_ = n_readers_;

    leader_->set_leader();

    for (i = 0; i < n_readers_; i++)
    {
	if (readers_[i] != leader_)
	    readers_[i]->set_leader(leader_);
	readers_[i]->start(playback_start_, base_offset_);
    }
}

void Writer::reader_finished(SSRC_Reader *s)
{
    n_readers_running_--;

    if (n_readers_running_ == 0)
    {
	readers_finished();
    }
}

void Writer::readers_finished()
{
    int i;
    
    printf("Writer:readers_finished\n");

    for (i = 0; i < n_clients_; i++)
    {
	send_callback(clients_[i], "stream_done", "done");
    }
    reactor_->cancel_timer(counter_id_);
    counter_id_ = 0;
    running_ = 0;
}

void Writer::stop()
{
    printf("Writer::stop()\n");

    reactor_->cancel_timer(counter_id_);
    counter_id_ = 0;

    running_ = 0;

    int i;
    
    for (i = 0; i < n_readers_; i++)
    {
	readers_[i]->stop();
    }
}

void Writer::reset()
{
    printf("Writer::reset()\n");

    int i;
    
    for (i = 0; i < n_readers_; i++)
    {
	readers_[i]->reset();
    }
}

void Writer::quit()
{
    printf("Writer::quit()\n");

    int i;
    
    for (i = 0; i < n_readers_; i++)
    {
	readers_[i]->stop();
    }
    finished_ = 1;
}

void Writer::send_rtp_packet(int rtp_type, char *packet, int len)
{
    int i;

    for (i = 0; i < n_destinations_; i++)
    {
	destinations_[i]->send_rtp_packet(rtp_type, packet, len);
    }
}

void Writer::send_rtcp_packet(char *packet, int len)
{
    int i;

    for (i = 0; i < n_destinations_; i++)
    {
	destinations_[i]->send_rtcp_packet(packet, len);
    }
}

void Writer::send_rtcp_packet(int rtp_type, char *packet, int len)
{
    int i;

    for (i = 0; i < n_destinations_; i++)
    {
	destinations_[i]->send_rtcp_packet(rtp_type, packet, len);
    }
}

void Writer::add_reader(SSRC_Reader *reader)
{
    readers_[n_readers_++] = reader;
}

void Writer::remove_reader(SSRC_Reader *reader)
{
    int i, idx;

    idx = -1;
    for (i = 0; i < n_readers_; i++)
    {
	if (readers_[i] == reader)
	{
	    idx = i;
	    break;
	}
    }

    if (idx >= 0)
    {
	for (i = idx; i < n_readers_- 1; i++)
	{
	    readers_[i] = readers_[i+1];
	}
	n_readers_--;
    }
}

void Writer::add_destination(Destination *dest)
{
    destinations_[n_destinations_++] = dest;
}

void Writer::remove_destination(Destination *destination)
{
    int i, idx;

    idx = -1;
    for (i = 0; i < n_destinations_; i++)
    {
	if (destinations_[i] == destination)
	{
	    idx = i;
	    break;
	}
    }

    if (idx >= 0)
    {
	for (i = idx; i < n_destinations_ - 1; i++)
	{
	    destinations_[i] = destinations_[i+1];
	}
	n_destinations_--;
    }
}

Destination *Writer::find_destination(char *host, int port)
{
    int i;

    for (i = 0; i < n_destinations_; i++)
    {
	if (destinations_[i]->matches(host, port))
	    return destinations_[i];
    }
    return NULL;
}

/*
 * Following are all the Nexus message handlers. They just decode the
 * message buffer and invoke the proper method on the Writer instance.
 */

void Writer::add_client_handler(void *address, nexus_buffer_t *buffer)
{
    nexus_global_pointer_t *gp;

    gp = new nexus_global_pointer_t;
    nexus_get_global_pointer(buffer, gp, 1);

    ((Writer *) address)->add_client(gp);
}

void Writer::add_file_handler(void *address, nexus_buffer_t *buffer)
{
    int len;
    char *file;
    int *packet_types, n_packet_types;

    nexus_get_int(buffer, &len, 1);
    file = new char[len + 1];
    nexus_get_char(buffer, file, len);
    file[len] = '\0';

    nexus_get_int(buffer, &n_packet_types, 1);
    packet_types = new int[n_packet_types];
    nexus_get_int(buffer, packet_types, n_packet_types);

    ((Writer *) address)->add_file(file, n_packet_types, packet_types);
}

void Writer::remove_file_handler(void *address, nexus_buffer_t *buffer)
{
    int len;
    char *file;

    nexus_get_int(buffer, &len, 1);
    file = new char[len + 1];
    nexus_get_char(buffer, file, len);
    file[len] = '\0';

    ((Writer *) address)->remove_file(file);
}

void Writer::add_destination_handler(void *address, nexus_buffer_t *buffer)
{
    int len;
    int port;
    char *host;
    nexus_global_pointer_t *cb_gp;
    int *packet_types, n_packet_types;
    
    nexus_get_int(buffer, &len, 1);
    host = new char[len + 1];
    nexus_get_char(buffer, host, len);
    host[len] = '\0';

    nexus_get_int(buffer, &port, 1);

    cb_gp = new nexus_global_pointer_t;
    nexus_get_global_pointer(buffer, cb_gp, 1);

    nexus_get_int(buffer, &n_packet_types, 1);
    packet_types = new int[n_packet_types];
    nexus_get_int(buffer, packet_types, n_packet_types);

    ((Writer *) address)->add_destination(host, port, cb_gp, 
					  n_packet_types, packet_types);
}

void Writer::remove_destination_handler(void *address, nexus_buffer_t *buffer)
{
    int len;
    int port;
    char *host;

    nexus_get_int(buffer, &len, 1);
    host = new char[len + 1];
    nexus_get_char(buffer, host, len);
    host[len] = '\0';

    nexus_get_int(buffer, &port, 1);

    ((Writer *) address)->remove_destination(host, port);
}

void Writer::start_handler(void *address, nexus_buffer_t *buffer)
{
    ((Writer *) address)->start();
}

void Writer::stop_handler(void *address, nexus_buffer_t *buffer)
{
    ((Writer *) address)->stop();
}

void Writer::reset_handler(void *address, nexus_buffer_t *buffer)
{
    ((Writer *) address)->reset();
}

void Writer::quit_handler(void *address, nexus_buffer_t *buffer)
{
    ((Writer *) address)->quit();
}

